#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>

char *proxyaddr;
char *proxyport;
char *proxyproto;
char *destaddr;
char *proxytag;

int verbose;
int timeout;

int nread;

char *progname;

static char buf[2048];

static void done(int code) {
  write(3, &nread, sizeof(nread));
  exit(code);
}

void error(int errnum, const char *fmt, ...) {
  va_list ap;
  char buf[128];
  int l = sprintf(buf, "%s: %s: ", proxytag, progname);
  va_start(ap, fmt); l += vsprintf(buf + l, fmt, ap); va_end(ap);
  if (errnum)
    l += sprintf(buf + l, ": %s\n", strerror(errnum));
  else
    buf[l++] = '\n';
  write(2, buf, l);
  done(1);
}

void cinfo(int level, const char *fmt, ...) {
  if (verbose >= level) {
    va_list ap;
    char buf[2048];
    int l = sprintf(buf, "%s: ", proxytag);
    va_start(ap, fmt);
    l += vsprintf(buf + l, fmt, ap);
    va_end(ap);
    buf[l++] = '\n';
    write(2, buf, l);
  }
}

void timedout(int sig) {
  cinfo(2, "timeout");
  done(1);
}

void senddata(const char *fmt, ...) {
  va_list ap;
  int l;
  va_start(ap, fmt);
  l = vsprintf(buf, fmt, ap);
  va_end(ap);
  cinfo(3, ">> %s", buf);
  buf[l++] = '\r';
  buf[l++] = '\n';
  if (write(0, buf, l) < 0) {
    cinfo(2, "write to remote: %s", strerror(errno));
    done(1);
  }
}

void waitfor(int e) {
  while(fgets(buf, 64, stdin)) {
    nread += strlen(buf); /*XX if we read \0's? */
    if (verbose > 2) {
      char *s = buf, *d = buf + 65;
      unsigned char c;
      while((c = (unsigned char)(*s++)) != 0) {
	if (c == '\r')
	  d += sprintf(d, "\\r");
	else if (c == '\n')
	  d += sprintf(d, "\\n");
	else if (c < ' ' || c >= 0177)
	  d += sprintf(d, "\\%o", c);
	else
	  *d++ = c;
      }
      *d = '\0';
      cinfo(3, "<< %s", buf + 65);
    }
    if (buf[3] == ' ' && atoi(buf) / 100 == e) {
      return;
    }
    if (nread > 512) {
      cinfo(2, "too much input, closing connection");
      done(1);
    }
  }
  cinfo(2, "connection closed");
  done(1);
}

#define COOKIE_PORT 200
#define COOKIE_LEN 32

char *getcookie(const char *cookie_host,
		const char *user, const char *pass) {
  struct sockaddr_in sin;
  int fd;
  int len;
  struct hostent *he;

  he = gethostbyname(cookie_host);
  if (!he)
    return 0;
  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(COOKIE_PORT);
  memcpy(&sin.sin_addr, he->h_addr_list[0], sizeof(sin.sin_addr));
  fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (fd < 0)
    return 0;
  if (connect(fd, (struct sockaddr*)&sin, sizeof(sin)) < 0)
    return 0;
  len = sprintf(buf, "%s\n%s\n", user, pass);
  if (send(fd, buf, len, 0) < 0 ||
      recv(fd, buf, COOKIE_LEN, 0) != COOKIE_LEN)
    return 0;
  close(fd);
  buf[COOKIE_LEN] = '\0';
  return strdup(buf);
}

int main(int argc, char **argv) {
  char buf[2048];
  int l;
  char *s;
  char *from, *rcpt;
  char *cookie_host, *user, *pass;

  if ((progname = strrchr(argv[0], '/')) == NULL)
    progname = argv[0];
  else
    argv[0] = ++progname;

  if (argc != 1 ||
      !(proxyaddr = getenv("PROXYADDR")) ||
      !(proxyport = getenv("PROXYPORT")) ||
      !(proxyproto = getenv("PROXYPROTO")) ||
      !(proxytag = getenv("PROXYTAG")) ||
      !(destaddr = getenv("DESTADDR"))) {
    fprintf(stderr, "%s: invalid usage\n", progname);
    return 2;
  }

  verbose = atoi(getenv("VERBOSE"));
  timeout = atoi(getenv("TIMEOUT"));

  signal(SIGALRM, timedout);
  alarm(timeout);

  if (!(from = getenv("DSBL_FROM")))
    from = "nobody";
  if (!(rcpt = getenv("DSBL_RCPT")))
    rcpt = "listme@listme.dsbl.org";

  senddata("HELO [%s]", proxyaddr);
  waitfor(2);
  senddata("MAIL FROM:<%s>", from);
  senddata("RCPT TO:<%s>", rcpt);
  senddata("DATA");
  waitfor(3);

  /* at this point, connection seems to be ok, so
   * double the timeout */
  alarm(timeout);

  if (!(cookie_host = getenv("DSBL_COOKIE_HOST")))
    cookie_host = "cookie.dsbl.org";
  if (!(user = getenv("DSBL_USER")))
    user = pass = "";
  else if (!(pass = getenv("DSBL_PASS")))
    pass = "";
  s = getcookie(cookie_host, user, pass);
  if (!s) {
    cinfo(0, "error getting cookie from %s, user \"%s\"; assuming open",
	  cookie_host, user);
    done(0);
  }

  cinfo(3, ">> sending %s", s);
  l = sprintf(buf,
"Message-ID: <%s@%s>\r\n"
"To: <%s>\r\n"
"Subject: Open %s Proxy test message\r\n"
"\r\n"
"DSBL LISTME: %s %s\r\n"
"%s\r\n"
"Port %s, connect to %s\r\n"
"DSBL END\r\n"
"\r\n"
".\r\n",
	      s, cookie_host, /* Message-ID */
	      rcpt, /* To: */
	      proxyproto, /* Subj */
	      proxyproto, proxyaddr, /* dsbl listme */
	      s, /* dsbl cookie */
	      proxyport, destaddr);
  if (write(0, buf, l) < 0) {
    cinfo(1, "error sending message (assuming open): %s", strerror(errno));
    done(0);
  }
  waitfor(2);
  write(0, "QUIT\r\n", 6);
  done(0);
  return 0;
}
