/* simple event library. 
 * Written by Michael Tokarev <mjt@corpit.ru>
 * LGPL
 */

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/select.h>
#include "event.h"

#ifndef __attribute__
#  ifdef __GNUC__
#  define __attribute__ __attribute__
# else
#  define __attribute__(x)
# endif
#endif
#ifndef UNUSED
# define UNUSED __attribute__((unused))
#endif

typedef struct event_timer {
  time_t when;			/* time when to trigger */
  event_callback *cbck;
  void *ctx;
  struct event_timer *next;	/* linked list: next timer */
} event_timer;

struct _event_ctx {
  int nfds;		/* current max active fd#-1 for select() */
  int nested;		/* flag to prevent nested calls of event_wait() */
  event_timer *timers;	/* list of timers sorted by time of trigger */
  event_timer *stimer;	/* a pointer to one spare timer structure */
  time_t when;		/* current event's time */
  fd_set xfds, wfds, rfds; /* ready-to-use fdsets consistent with io[] */
  struct {
    event_callback *cbck;
    void *ctx;
  } io[FD_SETSIZE]; /* array of fds indexed by fd */
};

static event_ctx event_default_ctx; /* zeroed */

#define set_ectx() if (!ectx) ectx = &event_default_ctx
#define def_ectx() (ectx ? ectx : &event_default_ctx)

event_ctx *event_ctx_new() {
  event_ctx *ectx = (event_ctx*)malloc(sizeof(event_ctx));
  if (!ectx) {
    errno = ENOMEM;
    return NULL;
  }
  memset(ectx, 0, sizeof(*ectx));
  return ectx;
}

void event_ctx_free(event_ctx *ectx) {
  set_ectx();
  while(ectx->timers) {
    event_timer *t = ectx->timers;
    ectx->timers = t->next;
    free(t);
  }
  if (ectx->stimer)
    free(ectx->stimer);
  if (ectx != &event_default_ctx)
    free(ectx);
  else
    memset(ectx, 0, sizeof(*ectx));
}

/* free a timer structure */
#define freetimer(t) \
  if (ectx->stimer) free(t); else ectx->stimer = t

/* request a new timer or cancel existing one */
int event_request_timer(event_ctx *ectx, int timeout,
                        event_callback *cbck, void *ctx, time_t now) {
  event_timer *t, **tp, *p;
  if (!cbck) {
    errno = EINVAL;
    return -1;
  }
  set_ectx();
  for(tp = &ectx->timers; (t = *tp) != NULL; tp = &t->next)
    if (t->cbck == cbck && t->ctx == ctx) {
      *tp = t->next;
      tp = &t->next;
      break;
    }
  if (timeout < 0) {
    if (!t) {
      errno = ENOENT;
      return -1;
    }
    freetimer(t);
  }
  else {
    if (!t) {
      if ((t = ectx->stimer) != NULL)
        ectx->stimer = NULL;
      else if ((t = (event_timer*)malloc(sizeof(event_timer))) == NULL) {
        errno = ENOMEM;
        return -1;
      }
      t->cbck = cbck;
      t->ctx = ctx;
    }
    t->when = timeout ? (now ? now : time(NULL)) + timeout : 0;
    tp = &ectx->timers;
    while((p = *tp) != NULL && p->when <= t->when)
      tp = &p->next;
    t->next = *tp;
    *tp = t;
  }
  return 0;
}

/* check if given fd is valid and return ret if not */
#define checkfdvalid(fd, ret) \
  set_ectx(); if (fd < 0 || fd >= FD_SETSIZE) { errno = EINVAL; return ret; }
/* check if I/O monitoring was requested fof given fd */
#define checkfdpresent(fd, ret) \
  if (!ectx->io[fd].cbck) { errno = ENOENT; return ret; }
#define checkfd(fd, ret) checkfdvalid(fd, ret) checkfdpresent(fd, ret)

/* request or cancel I/O monitoring */
int event_request_io(event_ctx *ectx, int fd, event_t event,
                     event_callback *cbck, void *ctx) {
  checkfdvalid(fd, -1);
  if (!cbck) { /* cancel request */
    checkfdpresent(fd, -1);
    FD_CLR(fd, &ectx->xfds);
    FD_CLR(fd, &ectx->rfds);
    FD_CLR(fd, &ectx->wfds);
    ectx->io[fd].ctx = ectx->io[fd].cbck = NULL;
    if (ectx->nfds <= fd + 1) {
      while(fd >= 0 && !ectx->io[fd].cbck)
        --fd;
      ectx->nfds = fd + 1;
    }
  }
  else { /* new monitoring request */
    if (event & EVENT_EXCPT) FD_SET(fd, &ectx->xfds);
    else FD_CLR(fd, &ectx->xfds);
    if (event & EVENT_READ) FD_SET(fd, &ectx->rfds);
    else FD_CLR(fd, &ectx->rfds);
    if (event & EVENT_WRITE) FD_SET(fd, &ectx->wfds);
    else FD_CLR(fd, &ectx->wfds);
    if (ectx->nfds <= fd) ectx->nfds = fd + 1;
    ectx->io[fd].cbck = cbck;
    ectx->io[fd].ctx = ctx;
  }
  return 0;
}

/* return a callback assotiated with fd */
event_callback *event_io_cbck(const event_ctx *ectx, int fd) {
  checkfd(fd, NULL);
  return ectx->io[fd].cbck;
}

/* return a context assotiated with fd */
void *event_io_ctx(const event_ctx *ectx, int fd) {
  checkfd(fd, NULL);
  return ectx->io[fd].ctx;
}

/* call a callback for a given fd */
int event_io_wakeup(event_ctx *ectx, int fd, event_t event) {
  checkfd(fd, -1);
  ectx->io[fd].cbck(ectx, fd, event, ectx->io[fd].ctx);
  return 0;
}

int event_events(const event_ctx *ectx) {
  set_ectx();
  return ectx->nfds || ectx->timers;
}

time_t event_time(const event_ctx *ectx) {
  return def_ectx()->when;
}

int event_wait(event_ctx *ectx, int timeout) {
  fd_set wfd, rfd, xfd;
  struct timeval tv, *tvp;
  event_timer *t;
  int n;
  int fd;
  time_t now;

  set_ectx();

  if (ectx->nested) {
    errno = EINPROGRESS;
    return -1;
  }

  if (!timeout)
    ;
  else if ((t = ectx->timers) != NULL) {
    if (!t->when)
      timeout = 0;
    else if ((now = t->when - time(NULL)) <= 0)
      timeout = 0;
    else if (timeout < 0 || (timeout > 0 && timeout < now))
      timeout = now;
  }
  if (timeout < 0)
    tvp = NULL;
  else {
    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    tvp = &tv;
  }

  rfd = ectx->rfds;
  wfd = ectx->wfds;
  xfd = ectx->xfds;

  n = select(ectx->nfds, &rfd, &wfd, &xfd, tvp);
  if (n < 0)
    return -1;

  ++ectx->nested;

  now = time(&ectx->when);

  while((t = ectx->timers) != NULL && now >= t->when) {
    ectx->timers = t->next;
    t->cbck(ectx, -1, EVENT_TIMER, t->ctx);
    freetimer(t);
  }

  for(fd = 0; n > 0 && fd < ectx->nfds; ++fd) {
    event_t e;
    if (!ectx->io[fd].cbck) continue;
    else if (FD_ISSET(fd, &xfd)) e = EVENT_EXCPT;
    else if (FD_ISSET(fd, &wfd)) e = EVENT_WRITE;
    else if (FD_ISSET(fd, &rfd)) e = EVENT_READ;
    else continue;
    ectx->io[fd].cbck(ectx, fd, e, ectx->io[fd].ctx);
    --n;
  }

  --ectx->nested;
  return 0;
}

#ifdef TEST

#include <stdio.h>

static void
ding(event_ctx *ectx, int UNUSED fd, event_t UNUSED event, void *ctx) {
  char *c = (char*)ctx;
  write(1, c, 1);
  if (*c >= 'A' && *c <= 'Z') *c += 'a' - 'A';
  else if (*c >= 'a' && *c <= 'z') *c += 'A' - 'a';
  event_request_timer(ectx, 2, ding, ctx, event_time(ectx));
}

static void
echo(event_ctx UNUSED *ectx, int UNUSED fd, event_t UNUSED event,
     void UNUSED *ctx) {
  char buf[BUFSIZ];
  if (fgets(buf, sizeof(buf), stdin) == 0)
    exit(0);
  printf("Result: %s", buf);
}

int main(void) {
  static char text[] = "\rdingdong ";
  char   *cp;

  for (cp = text; *cp; cp++)
     event_request_timer(NULL, 0, ding, cp, 0);
  event_request_io(NULL, fileno(stdin), EVENT_READ, echo, 0);
  for (;;)
    event_wait(NULL, -1);
}

#endif
